# -*- coding: utf-8  -*-
"""Package for tests."""

# Copyright (C) 2015 Alexander Jones
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import unittest


class TestCase(unittest.TestCase):

    """Subclass of unittest.TestCase."""

    if not hasattr(unittest.TestCase, 'assertCountEqual'):

        def assertCountEqual(self, *args, **kwargs):
            """Wrapper of assertItemsEqual()."""
            return self.assertItemsEqual(*args, **kwargs)



# -*- coding: utf-8  -*-
"""Round-robin scheduling."""

# Copyright (C) 2015 Alexander Jones
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals

import copy
import random

from . import NoMatchFound, ScheduleGenerationFailed
from .scheduler import Scheduler


class RoundRobinScheduler(Scheduler):

    """A standard round-robin scheduler."""

    def __init__(self, teams, meetings=0):
        """Constructor.

        @param teams: A list of teams or the number of teams
        @type teams: list or int
        @param meetings: The number of times teams meet each other
        @type meetings: int
        """
        if not isinstance(teams, list):
            teams = list(range(1, teams + 1))
        if len(teams) % 2 == 1:
            teams.append(None)
        self.teams = teams
        if meetings % 2 == 1:
            raise ValueError('Odd meeting numbers are not yet supported.')
        self.meetings = meetings

    @property
    def match_count(self):
        """The number of matches per round."""
        return int(len(self.teams) / 2)

    @property
    def round_count(self):
        """The number of rounds in a season."""
        return int((len(self.teams) - 1) * self.meetings)

    def generate_matches(self):
        """Generate the matches for the season.

        @return: The matches to schedule
        @rtype: list
        """
        is_odd = self.meetings % 2 == 1
        if is_odd:
            raise ValueError('Odd meeting numbers are not yet supported.')
            # evens = int((self.meetings - 1) / 2)
        else:
            evens = int(self.meetings / 2)

        # Even meetings
        matches = [(team, opp)
                   for team in self.teams
                   for opp in self.teams
                   if team != opp] * evens
        # TODO: Odd meetings
        return matches

    def generate_round(self, matches):
        """Generate a round.

        @param matches: The generated matches
        @type matches: list
        @return: The generated round
        @rtype: list
        """
        round = []
        try:
            random.shuffle(matches)
            round.append(matches.pop(0))
            poss = copy.copy(matches)
            for __ in range(1, self.match_count):
                match = Scheduler.find_unique_match(round, poss)
                round.append(match)
                matches.remove(match)
            return round
        except NoMatchFound:
            matches.extend(round)
            return None

    def generate_schedule(self, try_once=False):
        """Generate the schedule.

        @param try_once: Whether to only try once to generate a schedule
        @type try_once: bool
        @return: The generated schedule
        @rtype: list of lists of tuples
        @raise RuntimeError: Failed to create schedule within limits
        """
        rounds = []
        matches = self.generate_matches()

        for __ in range(self.round_count):
            for ___ in range(10):
                next_round = self.generate_round(matches)
                if next_round:
                    rounds.append(next_round)
                    break
            else:
                if not try_once:
                    return self.generate_schedule()
                else:
                    raise ScheduleGenerationFailed('Schedule generation failed.')

        return rounds


class DoubleRoundRobinScheduler(RoundRobinScheduler):

    """A standard double round-robin scheduler.

    This is an alias of RoundRobinScheduler, with meetings=2.
    """

    def __init__(self, teams):
        """Constructor.

        @param teams: A list of teams or the number of teams
        @type teams: list or int
        """
        super(DoubleRoundRobinScheduler, self).__init__(teams, meetings=2)


class QuadrupleRoundRobinScheduler(RoundRobinScheduler):

    """A standard quadruple round-robin scheduler.

    This is an alias of RoundRobinScheduler, with meetings=4.
    """

    def __init__(self, teams):
        """Constructor.

        @param teams: A list of teams or the number of teams
        @type teams: list or int
        """
        super(QuadrupleRoundRobinScheduler, self).__init__(teams, meetings=4)



# -*- coding: utf-8  -*-
"""Scheduler package."""

# Copyright (C) 2015 Alexander Jones
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals


class ScheduleGenerationFailed(RuntimeError):

    """Exception for failed schedule generation.

    This is raised by implementations of Scheduler.generate_schedule() in the
    event that it cannot generate a schedule. This is usually because try_once
    was passed and the first attempt failed. The method will otherwise call
    itself until a schedule can be generated, and this will only fail if the
    recursion limit is reached, when a different error will be raised.
    """

    pass


class NoMatchFound(RuntimeError):

    """Exception for failure to find suitable match.

    This exception is meant for internal use. If this exception reaches
    consumers, there is a coding error and this exception should not be
    caught.
    """

    pass



"""A setuptools based setup module."""

from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='competitions-scheduler',
    version='0.1',

    description='Generic schedulers for competitions',
    long_description=long_description,

    url='https://github.com/happy5214/competitions-scheduler',

    author='Alexander Jones',
    author_email='happy5214@gmail.com',

    license='LGPLv3+',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    keywords='competitions schedules',

    packages=find_packages(exclude=['docs', 'tests*']),

    test_suite='tests',
)



# -*- coding: utf-8  -*-
"""Base class code for schedulers."""

# Copyright (C) 2015 Alexander Jones
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from . import NoMatchFound


class Scheduler(object):

    """Base class for schedulers.

    This class includes general-purpose scheduler functionality.
    """

    @staticmethod
    def matches_share_opponents(match1, match2):
        """Check matches for shared opponents.

        @param match1: The first match
        @type match1: tuple
        @param match2: The second match
        @type match2: tuple
        @return: Whether the first and second matches share an opponent
        @rtype: bool
        """
        (match1_1, match1_2) = match1
        return (match1_1 in match2 or match1_2 in match2)

    @staticmethod
    def find_unique_match(matches, rest):
        """Find a unique match.

        @param matches: The matches to check against
        @type matches: list
        @param rest: The matches to check for a unique match
        @type rest: list
        @return: The first unique match found
        @rtype: tuple
        @raise NoMatchFound: If no unique match is found
        """
        while True:  # This will exit with the return statement or an error.
            try:
                possibility = rest.pop(0)  # Possibly unique match
                for match in matches:
                    if Scheduler.matches_share_opponents(match, possibility):
                        possibility = None  # Match is not unique.
                        break  # Try again.
                if possibility:  # No more matches to compare against.
                    return possibility  # This match is unique.
            except IndexError:  # We've run out of possibilities.
                raise NoMatchFound  # Raise exception to caller

    def generate_schedule(self, try_once=False):
        """Generate the schedule.

        This method must be overridden by subclasses.

        @param try_once: Whether to only try once to generate a schedule
        @type try_once: bool
        @return: The generated schedule
        @rtype: list of lists of tuples
        """
        raise NotImplementedError
